/*
 * gstripe.c
 *
 * Copyright (c) 2003 Pawel Jakub Dawidek <jules@garage.freebsd.pl>
 * All rights reserved.
 *
 * $Id: gstripe.c,v 1.3 2004/01/08 12:37:16 jules Exp $
 *
 */

#include <sys/param.h>
#include <sys/disk.h>
#include <sys/endian.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <assert.h>

#include "g_raid0.h"


enum { UNSET, MARK, INFO } action = UNSET;
static const char *comm;
static uint32_t stripesize = 262144;
static int verbose = 0;


static void
usage(void)
{

	fprintf(stderr, "usage: %s -m [-v] [-s stripesize] name device1 "
	    "[device2 [...]]\n", comm);
	fprintf(stderr, "       %s -i device1 [device2 [...]]\n", comm);
	exit(EXIT_FAILURE);
}

static void
disk_mark(int argc, char *argv[])
{
	struct g_raid0_metadata md;
	unsigned char *sector;
	unsigned sectorsize;
	off_t mediasize;
	int fd, status;
	ssize_t data;
	uint16_t i;

	if (argc < 2)
		usage();

	if (stripesize < 512 || !powerof2(stripesize))
		errx(EXIT_FAILURE, "Invalid stripesize.");
	md.md_id = arc4random();
	if (strlcpy(md.md_name, argv[0], sizeof(md.md_name)) >=
	    sizeof(md.md_name)) {
		errx(EXIT_FAILURE, "Name too long.");
	}
	strlcpy(md.md_magic, G_RAID0_MAGIC, sizeof(md.md_magic));
	argc--;
	argv++;

	md.md_all = argc;
	md.md_stripesize = stripesize;
	status = EXIT_SUCCESS;

	for (i = 0; i < argc; i++) {
		fd = open(argv[i], O_WRONLY);
		if (fd == -1) {
			fprintf(stderr, "Can't open %s: %s.\n", argv[i],
			    strerror(errno));
			status = EXIT_FAILURE;
			continue;
		}
		if (ioctl(fd, DIOCGSECTORSIZE, &sectorsize) < 0) {
			fprintf(stderr, "Can't get sector size for %s: %s.\n",
			    argv[i], strerror(errno));
			status = EXIT_FAILURE;
			continue;
		}
		if (ioctl(fd, DIOCGMEDIASIZE, &mediasize) < 0) {
			fprintf(stderr, "Can't get media size for %s: %s.\n",
			    argv[i], strerror(errno));
			status = EXIT_FAILURE;
			continue;
		}
		assert(sectorsize >= sizeof(md));
		if (stripesize < sectorsize) {
			fprintf(stderr, "Invalid stripesize for %s.\n",
			    argv[i]);
			status = EXIT_FAILURE;
			continue;
		}
		sector = malloc(sectorsize);
		if (sector == NULL) {
			fprintf(stderr, "Can't allocate memory for %s: %s.\n",
			    argv[i], strerror(errno));
			status = EXIT_FAILURE;
			continue;
		}
		md.md_no = i;
		g_raid0_metadata_enc(&md, sector);
		data = pwrite(fd, sector, sectorsize, mediasize - sectorsize);
		free(sector);
		close(fd);
		if (data != (ssize_t)sectorsize) {
			fprintf(stderr, "Problems while storing metadata on "
			    "%s: %s.\n", argv[i], strerror(errno));
			status = EXIT_FAILURE;
			continue;
		}
		if (verbose) {
			printf("Metadata stored on %s (id=%u).\n", argv[i],
			    (unsigned)md.md_id); 
		}
	}

	exit(status);
}

static void
disk_info(int argc, char *argv[])
{
	struct g_raid0_metadata md;
	unsigned char *sector;
	unsigned sectorsize;
	off_t mediasize;
	int fd, status;
	uint16_t i;

	if (argc < 1)
		usage();

	status = EXIT_SUCCESS;

	for (i = 0; i < argc; i++) {
		fd = open(argv[i], O_RDONLY);
		if (fd == -1) {
			fprintf(stderr, "Can't open %s: %s.\n", argv[i],
			    strerror(errno));
			status = EXIT_FAILURE;
			continue;
		}
		if (ioctl(fd, DIOCGSECTORSIZE, &sectorsize) < 0) {
			fprintf(stderr, "Can't get sector size for %s: %s.\n",
			    argv[i], strerror(errno));
			status = EXIT_FAILURE;
			continue;
		}
		if (ioctl(fd, DIOCGMEDIASIZE, &mediasize) < 0) {
			fprintf(stderr, "Can't get media size for %s: %s.\n",
			    argv[i], strerror(errno));
			status = EXIT_FAILURE;
			continue;
		}
		sector = malloc(sectorsize);
		if (sector == NULL) {
			fprintf(stderr, "Can't allocate memory for %s: %s.\n",
			    argv[i], strerror(errno));
			status = EXIT_FAILURE;
			continue;
		}
		if (pread(fd, sector, sectorsize, mediasize - sectorsize) !=
		    (ssize_t)sectorsize) {
			fprintf(stderr, "Problems while getting metadata from "
			    "%s: %s.\n", argv[i], strerror(errno));
			status = EXIT_FAILURE;
			free(sector);
			close(fd);
			continue;
		}
		close(fd);
		g_raid0_metadata_dec(sector, &md);
		free(sector);
		if (strcmp(md.md_magic, G_RAID0_MAGIC) != 0) {
			fprintf(stderr, "Not a GEOM_STRIPE disk (%s).\n", argv[i]);
			status = EXIT_FAILURE;
			continue;
		}
		printf("%s: name=%s, no=%u, all=%u, stripesize=%zu, id=%u\n",
		    argv[i], md.md_name, md.md_no, md.md_all, md.md_stripesize,
		    (unsigned)md.md_id);
	}

	exit(status);
}

int
main(int argc, char *argv[])
{
	int ch;

	comm = basename(argv[0]);

	while ((ch = getopt(argc, argv, "ims:v")) != -1) {
		switch (ch) {
		case 'i':
			if (action != UNSET)
				usage();
			action = INFO;
			break;
		case 'm':
			if (action != UNSET)
				usage();
			action = MARK;
			break;
		case 's':
			if (action != MARK)
				usage();
			stripesize = (uint32_t)strtoull(optarg, NULL, 0);
			break;
		case 'v':
			if (action != MARK)
				usage();
			verbose = 1;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	switch (action) {
	case MARK:
		disk_mark(argc, argv);
		/* NOTREACHED */
		break;
	case INFO:
		disk_info(argc, argv);
		/* NOTREACHED */
		break;
	default:
		usage();
	}

	exit(EXIT_SUCCESS);
}
