/*
 * g_raid0.c
 *
 * Copyright (c) 2003 Pawel Jakub Dawidek <jules@garage.freebsd.pl>
 * All rights reserved.
 *
 * $Id: raid0.c,v 1.1 2004/02/05 12:22:08 jules Exp $
 *
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bio.h>
#include <sys/endian.h>
#include <sys/sysctl.h>
#include <sys/malloc.h>
#include <geom/geom.h>

#include "g_stripe.h"
#include "g_raid0.h"


#define	G_RAID0_DEBUG(lvl, ...)	do {					\
	if (g_raid0_debug >= (lvl)) {					\
		printf("GEOM_" G_RAID0_CLASS_NAME "[%u]: ", lvl);	\
		printf(__VA_ARGS__);					\
		printf("\n");						\
	}								\
} while (0)

static MALLOC_DEFINE(M_RAID0, "raid0 data", "GEOM_" G_RAID0_CLASS_NAME " Data");

SYSCTL_NODE(_kern_geom, OID_AUTO, raid0, CTLFLAG_RW, 0,
    "GEOM_" G_RAID0_CLASS_NAME " stuff");
static u_int g_raid0_debug = 1;
SYSCTL_UINT(_kern_geom_raid0, OID_AUTO, debug, CTLFLAG_RW, &g_raid0_debug, 0,
    "Debug level");

static g_taste_t g_raid0_taste;

struct g_class g_raid0_class = {
	.name = G_RAID0_CLASS_NAME,
	.taste = g_raid0_taste,
};


static void
g_raid0_null_orphan(struct g_consumer *cp __unused)
{

	/* Do nothing, it is only used for tasting. */
}

static void
g_raid0_orphan(struct g_provider *pp)
{

	G_RAID0_DEBUG(1, "Removed provider %s from %s.", pp->name,
	    pp->geom->name);
}

static struct g_geom *
g_raid0_taste(struct g_class *mp, struct g_provider *pp, int flags __unused)
{
	struct g_raid0_metadata md;
	struct g_stripe_softc *sc;
	struct g_consumer *cp;
	struct g_geom *gp;
	u_char *buf, name[32];
	int error;

	g_trace(G_T_TOPOLOGY, "g_raid0_taste(%s, %s)", mp->name, pp->name);
	g_topology_assert();

	/*
	 * Create temporary geom and consumer to read first sector from disk.
	 */
	gp = g_new_geomf(mp, "raid0:taste");

	gp->orphan = g_raid0_null_orphan;
	cp = g_new_consumer(gp);
	g_attach(cp, pp);
	error = g_access(cp, 1, 0, 0);
	if (error != 0) {
		g_wither_geom(gp, ENXIO);
		return (NULL);
	}

	g_topology_unlock();
	buf = g_read_data(cp, pp->mediasize - pp->sectorsize, pp->sectorsize,
	    &error);
	g_topology_lock();
	g_access(cp, -1, 0, 0);
	g_wither_geom(gp, ENXIO);

	if (buf == NULL)
		return (NULL);
	if (error != 0) {
		g_free(buf);
		return (NULL);
	}

	g_raid0_metadata_dec(buf, &md);
	g_free(buf);

	if (strcmp(md.md_magic, G_RAID0_MAGIC) != 0)
		return (NULL);

	snprintf(name, sizeof(name), "%s.raid0", md.md_name);
	/*
	 * First we need to check if this belongs to this stripe.
	 */
	LIST_FOREACH(gp, &mp->geom, geom) {
		sc = gp->softc;
		if (sc == NULL)
			continue;
		if (strcmp(gp->name, name) != 0)
			continue;
		if (md.md_id != (uint32_t)sc->sc_priv)
			continue;
		break;
	}
	G_RAID0_DEBUG(2, "Metadata for %s: no=%u, all=%u, stripesize=%zu.",
	    md.md_name, md.md_no, md.md_all, md.md_stripesize);
	if (md.md_stripesize < pp->sectorsize || !powerof2(md.md_stripesize)) {
		G_RAID0_DEBUG(0, "Invalid stripesize for disk %s for %s.",
		    pp->name, name);
		return (NULL);
	}
	if (gp != NULL) {
		G_RAID0_DEBUG(1, "Adding provider %s to %s.", pp->name,
		    gp->name);
		error = g_stripe_attach(gp, pp, md.md_no, 1);
		if (error != 0) {
			G_RAID0_DEBUG(0, "Cannot add provider %s to %s.",
			    pp->name, gp->name);
			return (NULL);
		}
		g_stripe_add_slice(gp, pp, 0, pp->mediasize - pp->sectorsize);
		G_RAID0_DEBUG(0, "Added provider %s to %s.", pp->name,
		    gp->name);
	} else {
		G_RAID0_DEBUG(1, "Creating device %s.", name);
		gp = g_stripe_create(mp, name, md.md_all, md.md_stripesize,
		    pp->sectorsize, (void *)md.md_id, g_raid0_orphan);
		if (gp == NULL) {
			G_RAID0_DEBUG(0, "Cannot create device %s.", name);
			return (NULL);
		}
		G_RAID0_DEBUG(0, "Created device %s.", gp->name);
		sc = gp->softc;
		G_RAID0_DEBUG(1, "Adding provider %s to %s.", pp->name,
		    gp->name);
		error = g_stripe_attach(gp, pp, md.md_no, 1);
		if (error != 0) {
			G_RAID0_DEBUG(0, "Cannot add provider %s to %s.",
			    pp->name, gp->name);
			g_stripe_destroy(gp, 1);
			return (NULL);
		}
		g_stripe_add_slice(gp, pp, 0, pp->mediasize - pp->sectorsize);
		G_RAID0_DEBUG(0, "Added provider %s to %s.", pp->name,
		    gp->name);
	}

	return (gp);
}

DECLARE_GEOM_CLASS(g_raid0_class, g_raid0);
