/*
 * g_stripe.c
 *
 * Copyright (c) 2003 Pawel Jakub Dawidek <jules@garage.freebsd.pl>
 * All rights reserved.
 *
 * $Id: geom_stripe.c,v 1.1.1.1 2004/02/05 11:53:46 nick Exp $
 *
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bio.h>
#include <sys/endian.h>
#include <sys/sysctl.h>
#include <sys/malloc.h>
#include <geom/geom.h>

#include "g_stripe.h"


static MALLOC_DEFINE(M_STRIPE, "geom_stripe data", "GEOM_STRIPE allocations");

SYSCTL_NODE(_kern_geom, OID_AUTO, stripe, CTLFLAG_RW, 0, "GEOM_STRIPE stuff");
static u_int g_stripe_debug = 1;
SYSCTL_UINT(_kern_geom_stripe, OID_AUTO, debug, CTLFLAG_RW, &g_stripe_debug, 0,
    "Debug level");

#define	OPENED(who)	((who)->acr > 0 || (who)->acw > 0 || (who)->ace)

static int16_t
g_stripe_find_disk(struct g_stripe_softc *sc, struct g_provider *pp)
{
	register uint16_t i;

	for (i = 0; i < sc->sc_ndisks; i++) {
		if (sc->sc_disks[i].d_consumer == NULL)
			continue;
		if (sc->sc_disks[i].d_consumer->provider == pp)
			return (i);
	}

	return (-1);
}

static void
g_stripe_calc_groups(struct g_stripe_softc *sc)
{
	off_t diskend, offset;
	u_int ngroup;

	/*
	 * Probably too much, but don't worry, it will be truncated at the end.
	 */
	sc->sc_groups = malloc(sizeof(struct g_stripe_group) * sc->sc_ndisks,
	    M_STRIPE, M_WAITOK);

	offset = 0;
	diskend = 0;
	for (ngroup = 0; ngroup < sc->sc_ndisks; ngroup++) {
		struct g_stripe_group *group;
		off_t minsize = -1;
		u_int no;

		group = &sc->sc_groups[ngroup];
		group->g_disks =
		    malloc(sizeof(struct g_stripe_disk *) * sc->sc_ndisks,
		    M_STRIPE, M_WAITOK);
		group->g_ndisks = 0;

		for (no = 0; no < sc->sc_ndisks; no++) {
			struct g_stripe_disk *disk;

			disk = &sc->sc_disks[no];

			if (disk->d_size <= offset)
				continue;
			if (minsize == -1 || disk->d_size < minsize)
				minsize = disk->d_size;

			group->g_disks[group->g_ndisks] = disk;
			group->g_ndisks++;
		}
		if (minsize == -1) {
			/* No more groups. */
			free(group->g_disks, M_STRIPE);
			break;
		}
		group->g_disks =
		    realloc(group->g_disks,
		    sizeof(struct g_stripe_disk *) * group->g_ndisks, M_STRIPE,
		    M_WAITOK);
		if (ngroup == 0)
			group->g_start_offset = 0;
		else {
			group->g_start_offset =
			    sc->sc_groups[ngroup - 1].g_end_offset;
		}
		group->g_start_disk = diskend;
		group->g_end_offset = group->g_start_offset +
		    (minsize - offset) * group->g_ndisks;
		diskend += (minsize - offset);
		G_STRIPE_DEBUG(3, "%s: group %d: start_offset=%jd, "
		    "end_offset=%jd, start_disk=%jd.", sc->sc_provider->name,
		    ngroup, (intmax_t)group->g_start_offset,
		    (intmax_t)group->g_end_offset,
		    (intmax_t)group->g_start_disk);
		if (g_stripe_debug >= 3) {
			printf("%s: group %d disks:", sc->sc_provider->name,
			    ngroup);
			for (no = 0; no < group->g_ndisks; no++) {
				struct g_stripe_disk *disk;

				disk = group->g_disks[no];
				printf(" %s", disk->d_consumer->provider->name);
			}
			printf("\n");
		}
		offset = minsize;
	}
	sc->sc_groups = realloc(sc->sc_groups,
	    sizeof(struct g_stripe_group) * ngroup, M_STRIPE, M_WAITOK);
	sc->sc_ngroups = ngroup;
}

static void
g_stripe_check_and_run(struct g_geom *gp)
{
	struct g_stripe_disk *disk;
	struct g_stripe_softc *sc;
	struct g_provider *pp;
	uint16_t i;

	sc = gp->softc;
	KASSERT(sc != NULL, ("NULL softc in %s (class=%s, name=%s).",
	    __func__, gp->class->name, gp->name));

	/*
	 * Run provider if all slices were defined and
	 * all disks were introduced.
	 */
	pp = sc->sc_provider;
	if (pp == NULL) {
		off_t mindisksize;
		off_t totalsize = 0;

		mindisksize = -1;
		for (i = 0; i < sc->sc_ndisks; i++) {
			disk = &sc->sc_disks[i];
			if (disk->d_slices == NULL ||
			    disk->d_nslices != disk->d_ndone) {
				return;
			}
			if (disk->d_consumer == NULL)
				return;
			if (mindisksize == -1 || disk->d_size < mindisksize)
				mindisksize = disk->d_size;
			totalsize += disk->d_size;
		}

		KASSERT(mindisksize > 0,
		    ("No active disks? (mindisksize=%jd) (class=%s, name=%s).",
		    (intmax_t)mindisksize, gp->class->name, gp->name));

		g_stripe_calc_groups(sc);

		pp->mediasize = totalsize;
		pp->sectorsize = sc->sc_sectorsize;
	} else {
		if (sc->sc_nactive != sc->sc_ndisks) {
			/*
			 * Not all disks are reconnected.
			 */
			return;
		}
		/*
		 * Disk was disconnected and now it is back,
		 * activate provider.
		 */
	}
	G_STRIPE_DEBUG(1, "Provider %s activated.", pp->name);
	g_error_provider(pp, 0);
}

static void
g_stripe_orphan(struct g_consumer *cp)
{
	struct g_stripe_softc *sc;
	struct g_provider *pp;
	struct g_geom *gp;
	int16_t no;

	g_topology_assert();
	gp = cp->geom;
	sc = gp->softc;
	if (sc == NULL)
		return;

	pp = cp->provider;
	no = g_stripe_find_disk(sc, pp);
	KASSERT(no != -1,
	    ("Unknown consumer orphan (class=%s, name=%s, provider=%s).",
	    gp->class->name, gp->name, pp->name));

	g_error_provider(sc->sc_provider, ENXIO);
	sc->sc_disks[no].d_consumer = NULL;
	KASSERT(sc->sc_nactive > 0, ("nactive == 0? (class=%s, name=%s).",
	    gp->class->name, gp->name));
	sc->sc_nactive--;
	G_STRIPE_DEBUG(1, "Disk %s removed from %s.", pp->name, gp->name);
	if (OPENED(cp))
		g_access_rel(cp, -cp->acr, -cp->acw, -cp->ace);
	g_detach(cp);
	g_destroy_consumer(cp);
	/*
	 * Inform about orphan.
	 */
	if (sc->sc_orphan != NULL)
		sc->sc_orphan(pp);

	if (sc->sc_nactive > 0)
		return;

	G_STRIPE_DEBUG(1, "No active disks left in %s, removing device.",
	    gp->name);
	gp->class->destroy_geom(NULL, gp->class, gp);
}

/*
 * Find a group of disks where offset is placed.
 */
static struct g_stripe_group *
g_stripe_get_group(struct g_stripe_softc *sc, off_t offset)
{
	struct g_stripe_group *group;
	u_int no;

	for (no = 0; no < sc->sc_ngroups; no++) {
		group = &sc->sc_groups[no];
		if (offset < group->g_end_offset)
			return (group);
	}

	KASSERT(1 == 0, ("NULL group for %jd (class=%s, name=%s).",
	    (intmax_t)offset, sc->sc_provider->geom->class->name,
	    sc->sc_provider->geom->name));

	return (NULL);
}

/*
 * Send a request to given disk.
 * Here all calculations related to disk-slices are done.
 */
static int
g_stripe_send_request(struct g_stripe_disk *disk, struct bio *bp, off_t offset,
    off_t length, u_char *data)
{
	struct g_stripe_slice *slice;
	off_t end, diff, off, len;
	struct bio *cbp;
	uint16_t i;

	end = offset + length;

	for (i = 0; i < disk->d_nslices; i++) {
		slice = &disk->d_slices[i];
		if (slice->s_virt_end <= offset)
			continue;
		if (slice->s_virt_offset >= end)
			return (0);

		diff = offset - slice->s_virt_offset;
		off = slice->s_offset + diff;
		len = MIN(length, slice->s_length - diff);
		length -= len;
		offset += len;

		cbp = g_clone_bio(bp);
		if (cbp == NULL)
			return (ENOMEM);
		/*
		 * Fill in the component buf structure.
		 */
		cbp->bio_done = g_std_done;
		cbp->bio_offset = off;
		cbp->bio_data = data;
		data += len;
		cbp->bio_length = len;
		cbp->bio_to = disk->d_consumer->provider;
		G_STRIPE_LOGREQ(cbp, "Sending request. ");
		g_io_request(cbp, disk->d_consumer);

		if (length == 0)
			return (0);
	}

	KASSERT(length == 0,
	    ("Length is still greater than 0 (class=%s, name=%s).",
	    bp->bio_to->geom->class->name, bp->bio_to->geom->name));

	return (0);
}

/*
 * Find a disk for given offset and send request to this disk.
 * After sending request update offset, length and data arguments.
 */
static int
g_stripe_one_piece(struct g_stripe_softc *sc, struct bio *bp, off_t *offset,
    off_t *length, u_char **data)
{
	struct g_stripe_group *group;
	struct g_stripe_disk *disk;
	uint16_t ndisk, nstripe;
	off_t off, len, start;
	int error;

	off = *offset;
	len = *length;

	/*
	 * First we need to find a group of disks that are candidates
	 * to get this request.
	 */
	group = g_stripe_get_group(sc, off);

	/* We need to find a disk in this group, so we need to cut the rest. */
	off -= group->g_start_offset;

	/* Stripe number. */
	/* nstripe = off / stripesize; */
	nstripe = off >> (off_t)sc->sc_stripebits;
	/* Disk number. */
	ndisk = nstripe % group->g_ndisks;
	/* Start position in stripe. */
	/* start = off % sc->sc_stripesize; */
	start = off & (sc->sc_stripesize - 1);
	/* Start position in disk. */
	/* off = (nstripe / group->g_ndisks) * sc->sc_stripesize + start; */
	off = ((nstripe / group->g_ndisks) << sc->sc_stripebits) + start;
	off += group->g_start_disk;
	/* Length of data to operate. */
	len = MIN(len, sc->sc_stripesize - start);

	disk = group->g_disks[ndisk];
	error = g_stripe_send_request(disk, bp, off, len, *data);

	*offset += len;
	*length -= len;
	*data += len;

	return (error);
}

static void
g_stripe_start(struct bio *bp)
{
	struct g_geom *gp;
	struct g_stripe_softc *sc;
	off_t offset, length;
	u_char *data;
	int error;

	gp = bp->bio_to->geom;
	sc = gp->softc;
	KASSERT(sc != NULL, ("NULL softc in %s (class=%s, name=%s).",
	    __func__, gp->class->name, gp->name));

	/*
	 * If we don't have all disks connected, provider error
	 * should be set and g_stripe_start() should not be
	 * called at all.
	 */
	KASSERT(sc->sc_nactive == sc->sc_ndisks,
	    ("Not all disks connected, but %s was called (class=%s, name=%s).",
	    __func__, gp->class->name, gp->name));

	G_STRIPE_LOGREQ(bp, "Request received. ");

	switch (bp->bio_cmd) {
	case BIO_READ:
	case BIO_WRITE:
	case BIO_DELETE:
		/*
		 * Only those requests are supported.
		 */
		break;
	case BIO_GETATTR:
	default:
		g_io_deliver(bp, EOPNOTSUPP);
		return;
	}

	data = bp->bio_data;
	offset = bp->bio_offset;
	length = bp->bio_length;

	do {
		error = g_stripe_one_piece(sc, bp, &offset, &length, &data);
	} while (error == 0 && length > 0);

	if (error != 0) {
		/*
		 * Deny remaining part. This is pointless
		 * to split rest of the request, bacause
		 * we're setting bio_error here, so all
		 * request will be denied, anyway.
		 */
		bp->bio_completed += length;
		if (bp->bio_error == 0)
			bp->bio_error = error;
		if (bp->bio_completed == bp->bio_length)
			g_io_deliver(bp, bp->bio_error);
		return;
	}
	KASSERT(length == 0, ("Length != 0 (%jd) in %s (class=%s, name=%s).",
	    (intmax_t)length, __func__, gp->class->name, gp->name));
}

static int
g_stripe_destroy_geom(struct gctl_req *req __unused, struct g_class *mp,
     struct g_geom *gp)
{
	struct g_stripe_softc *sc;
	struct g_provider *pp;
	uint16_t i;

	sc = gp->softc;
	if (sc == NULL)
		return (0);

	pp = sc->sc_provider;
	if (mp == NULL) {
		/*
		 * If 'mp' is set to NULL, it means, that we were called from
		 * g_stripe_destroy(), so we can assume that provider is not
		 * open or we want force mode.
		 */
	} else {
		if (OPENED(pp))
			return (EBUSY);
	}

	G_STRIPE_DEBUG(1, "Destroyed device %s.", gp->name);

	g_error_provider(sc->sc_provider, ENXIO);
	gp->softc = NULL;
	g_wither_geom(gp, ENXIO);
	for (i = 0; i < sc->sc_ngroups; i++)
		free(sc->sc_groups[i].g_disks, M_STRIPE);
	free(sc->sc_groups, M_STRIPE);
	for (i = 0; i < sc->sc_ndisks; i++)
		free(sc->sc_disks[i].d_slices, M_STRIPE);
	free(sc->sc_disks, M_STRIPE);
	free(sc, M_STRIPE);

	return (0);
}

static int
g_stripe_access(struct g_provider *pp, int dr, int dw, int de)
{
	struct g_consumer *cp1, *cp2;
	struct g_stripe_softc *sc;
	struct g_geom *gp;
	int error;

	gp = pp->geom;
	sc = gp->softc;

	if (sc == NULL) {
		/*
		 * It looks like geom is being withered.
		 * In that case we allow only negative requests.
		 */
		KASSERT(dr <= 0 && dw <= 0 && de <= 0,
		    ("Positive access request (class=%s, name=%s).",
		    gp->class->name, gp->name));
		return (0);
	}
	if (dr > 0 || dw > 0 || de > 0) {
		/*
		 * If we don't have all disks connected, provider error
		 * should be set and g_stripe_access() should not be
		 * called at all.
		 */
		KASSERT(sc->sc_nactive == sc->sc_ndisks,
		    ("Not all disks connected, but %s was called (class=%s, name=%s).",
		    __func__, gp->class->name, gp->name));
	}

	/* On first open, grab an extra "exclusive" bit */
	if (!OPENED(pp));
		de++;
	/* ... and let go of it on last close. */
	if ((pp->acr + dr) == 0 && (pp->acw + dw) == 0 && (pp->ace + de) == 1)
		de--;

	error = ENXIO;
	LIST_FOREACH(cp1, &gp->consumer, consumer) {
		error = g_access_rel(cp1, dr, dw, de);
		if (error == 0)
			continue;
		/*
		 * If we fail here, backout all previous changes.
		 */
		LIST_FOREACH(cp2, &gp->consumer, consumer) {
			if (cp1 == cp2)
				return (error);
			g_access_rel(cp2, -dr, -dw, -de);
		}
		/* NOTREACHED */
	}

	return (0);
}

/*
 * Create device.
 */
struct g_geom *
g_stripe_create(struct g_class *mp, const char *name, uint16_t ndisks,
    uint32_t stripesize, uint32_t sectorsize, void *priv,
    void (*orphan)(struct g_provider *))
{
	struct g_stripe_softc *sc;
	struct g_geom *gp;
	struct g_provider *pp;
	uint16_t i;

	G_STRIPE_DEBUG(1, "Creating device %s.", name);

	KASSERT(stripesize >= 512,
	    ("Too small stripe size: %zu (512 is minimum) (class=%s, name=%s).",
	    stripesize, mp->name, name));
	KASSERT(powerof2(stripesize),
	    ("Stripe size should be power-of-2, %zu is not (class=%s, name=%s).",
	    stripesize, mp->name, name));
	KASSERT(ndisks > 1,
	    ("Too few disks: %d (2 is minimum) (class=%s, name=%s).",
	    ndisks, mp->name, name));

	gp = g_new_geomf(mp, "%s", name);
	if (gp == NULL)
		return (NULL);

	if (mp->destroy_geom == NULL)
		mp->destroy_geom = g_stripe_destroy_geom;
	gp->start = g_stripe_start;
	gp->spoiled = g_stripe_orphan;
	gp->orphan = g_stripe_orphan;
	gp->access= g_stripe_access;

	sc = malloc(sizeof(*sc), M_STRIPE, M_WAITOK | M_ZERO);

	sc->sc_disks = malloc(sizeof(struct g_stripe_disk) * ndisks, M_STRIPE,
	    M_WAITOK | M_ZERO);
	sc->sc_ndisks = ndisks;
	sc->sc_nactive = 0;
	for (i = 0; i < ndisks; i++) {
		sc->sc_disks[i].d_consumer = NULL;
		sc->sc_disks[i].d_size = 0;
		sc->sc_disks[i].d_slices = NULL;
		sc->sc_disks[i].d_nslices = 0;
		sc->sc_disks[i].d_ndone = 0;
	}

	sc->sc_sectorsize = sectorsize;
	sc->sc_stripesize = stripesize;
	sc->sc_stripebits = BITCOUNT(sc->sc_stripesize - 1);
	sc->sc_provider = NULL; /* We don't know mediasize yet. */
	sc->sc_groups = NULL;
	sc->sc_ngroups = 0;
	sc->sc_orphan = orphan;
	sc->sc_priv = priv;
	gp->softc = sc;

	if (mp->destroy_geom == NULL) {
		/*
		 * This should be only used if caller don't want
		 * dispose sc_priv before geom is destroyed.
		 */
		mp->destroy_geom = g_stripe_destroy_geom;
	}

	pp = g_new_providerf(gp, "%s", gp->name);
	pp->mediasize = 0;	/* Not yet. */
	pp->sectorsize = 0;	/* Not yet. */
	sc->sc_provider = pp;
	/* Leave provider with error for now. */

	return (gp);
}

/*
 * Add usable slice inside given provider.
 */
void
g_stripe_add_slice(struct g_geom *gp, struct g_provider *pp, off_t offset,
    off_t length)
{
	struct g_stripe_softc *sc;
	struct g_stripe_disk *disk;
	off_t disksize, prevend;
	int16_t no;

	sc = gp->softc;
	KASSERT(sc != NULL, ("NULL softc in %s (class=%s, name=%s).",
	    __func__, gp->class->name, gp->name));

	no = g_stripe_find_disk(sc, pp);
	KASSERT(no != -1, ("Unknown disk: %s (class=%s, name=%s).",
	    pp->name, gp->class->name, gp->name));

	disk = &sc->sc_disks[no];
	KASSERT(disk->d_slices != NULL,
	    ("Adding slice before disk? (class=%s, name=%s, provider=%s).",
	    gp->class->name, gp->name, pp->name));
	KASSERT(disk->d_ndone < disk->d_nslices,
	    ("Too many slices introduced (class=%s, name=%s, disk=%s).",
	    gp->class->name, gp->name, pp->name));

	no = disk->d_ndone;
	if (no == 0)
		prevend = 0;
	else {
		prevend = disk->d_slices[no - 1].s_offset +
		    disk->d_slices[no - 1].s_length;
		/*
		 * End of previous slice has to be less than start of
		 * the next slice. Yes, less than, it can't be equal, because
		 * those two slices could be one then.
		 */
		KASSERT(prevend < offset,
		    ("Slices overlap (class=%s, name=%s, disk=%s).",
		    gp->class->name, gp->name, pp->name));
	}

	disk->d_slices[no].s_virt_offset = prevend;
	disk->d_slices[no].s_virt_end = prevend + length;
	disk->d_slices[no].s_offset = offset;
	disk->d_slices[no].s_length = length;

	disk->d_ndone++;

	if (disk->d_nslices != disk->d_ndone)
		return;

	/* Ok, disk complete. */

	G_STRIPE_DEBUG(1, "All slices defined for %s.", pp->name);

	/*
	 * Disk size should be multiple of stripe size.
	 */
	disksize = (disk->d_slices[no].s_virt_end & ~(sc->sc_stripesize - 1));
	KASSERT(disksize > 0, ("Size of disk %s is 0 (class=%s, name=%s).",
	    pp->name, gp->class->name, gp->name));
	if (sc->sc_provider != NULL) {
		/*
		 * It looks like disk was disconnect and is connected
		 * again. Be sure that the same disk was reconnected
		 * by checking its currect and previous size.
		 */

		KASSERT(disk->d_size == disksize,
		    ("Different disk reconnected! (%jd != %jd) (class=%s, name=%s, disk=%s).",
		    (intmax_t)disk->d_size, (intmax_t)disksize,
		    gp->class->name, gp->name, pp->name));
	}
	disk->d_size = disksize;
	printf("DISKSIZE: %s %jd\n", disk->d_consumer->provider->name, (intmax_t)disksize);

	sc->sc_nactive++;
	KASSERT(sc->sc_nactive <= sc->sc_ndisks,
	    ("Too big nactive counter: %u (ndisks=%u)(class=%s, name=%s).",
	    sc->sc_nactive, sc->sc_ndisks, gp->class->name, gp->name));

	/*
	 * Check if that's all.
	 */
	g_stripe_check_and_run(gp);
}

/*
 * Attach given provider to given device.
 */
int
g_stripe_attach(struct g_geom *gp, struct g_provider *pp, uint16_t no,
    uint16_t nslices)
{
	struct g_stripe_softc *sc;
	struct g_stripe_disk *disk;
	struct g_provider *ourpp;
	struct g_consumer *cp;
	int error;

	sc = gp->softc;
	KASSERT(sc != NULL, ("NULL softc in %s (class=%s, name=%s).",
	    __func__, gp->class->name, gp->name));
	KASSERT(no < sc->sc_ndisks,
	    ("Invalid disk number: %u (class=%s, name=%s).",
	    no, gp->class->name, gp->name));
	KASSERT(sc->sc_disks[no].d_consumer == NULL,
	    ("Disk %u already initialized (class=%s, name=%s).",
	    no, gp->class->name, gp->name));
	KASSERT(nslices > 0, ("0 slices? (class=%s, name=%s, provider=%s).",
	    gp->class->name, gp->name, pp->name));

	cp = g_new_consumer(gp);
	if (cp == NULL)
		return (ENOMEM);	/* XXX: EWHAT? */
	error = g_attach(cp, pp);
	if (error != 0) {
		g_destroy_consumer(cp);
		return (error);
	}
	ourpp = sc->sc_provider;
	if (OPENED(ourpp)) {
		/*
		 * Other disks are already opened,
		 * so we need to open this disk as well.
		 */
		error = g_access(cp, ourpp->acr, ourpp->acw, ourpp->ace);
		if (error != 0) {
			/* I don't know what better can we do here. */
			g_detach(cp);
			g_destroy_consumer(cp);
			return (error);
		}
	}

	disk = &sc->sc_disks[no];
	disk->d_slices = malloc(sizeof(struct g_stripe_slice) * nslices,
	    M_STRIPE, M_WAITOK | M_ZERO);
	disk->d_nslices = nslices;
	disk->d_ndone = 0;
	disk->d_consumer = cp;

	return (0);
}

/*
 * Detach given provider from given device.
 */
int
g_stripe_detach(struct g_geom *gp, struct g_provider *pp)
{
	struct g_stripe_softc *sc;
	struct g_consumer *cp;
	int16_t no;

	sc = gp->softc;
	KASSERT(sc != NULL, ("NULL softc in %s (class=%s, name=%s).",
	    __func__, gp->class->name, gp->name));
	no = g_stripe_find_disk(sc, pp);
	if (no < 0) {
		/*
		 * Looks like provider was orphaned already.
		 */
		return (ENOENT);
	}

	g_stripe_orphan(cp);

	return (0);
}

/*
 * Destroy device.
 */
int
g_stripe_destroy(struct g_geom *gp, int force)
{
	struct g_stripe_softc *sc;

	sc = gp->softc;
	if (sc == NULL)
		return (0);

	if (OPENED(sc->sc_provider)) {
		if (!force)
			return (EBUSY);
	}

	return (g_stripe_destroy_geom(NULL, NULL, gp));
}
