/*
 * g_stripe.h
 *
 * (c) 2003 Pawel Jakub Dawidek <jules@garage.freebsd.pl>
 * All rights reserved.
 *
 * $Id: geom_stripe.h,v 1.1.1.1 2004/02/05 11:53:46 nick Exp $
 */

#ifndef	_G_STRIPE_H_
#define	_G_STRIPE_H_

#define	G_STRIPE_DEBUG(lvl, ...)		do {			\
	if (g_stripe_debug >= (lvl)) {					\
		printf("GEOM_STRIPE[%u]: ", lvl);			\
		printf(__VA_ARGS__);					\
		printf("\n");						\
	}								\
} while (0)
#define	G_STRIPE_LOGREQ(bp, ...)	do {				\
	if (g_stripe_debug >= 2) {					\
		printf("GEOM_STRIPE[2]: ");				\
		printf(__VA_ARGS__);					\
		g_print_bio(bp);					\
		printf("\n");						\
	}								\
} while (0)

extern struct sysctl_oid_list sysctl__kern_geom_children;

/*
 * Structure g_stripe_slice describe single slice on single disk.
 * Disk can contains many slices.
 *
 *      DATA     OFFSET
 * +-----------+ 0
 * | meta data | <- should not be touched
 * +-----------+ 1024
 * |           |
 * |  slice 0  |
 * |           |
 * +-----------+ 3072
 * | meta data | <- should not be touched
 * +-----------+ 4096
 * |           |
 * |           |
 * |  slice 1  |
 * |           |
 * |           |
 * +-----------+ 7168
 *
 * This disk have two slice:
 *
 * d_slices[0].s_offset = 1024
 * d_slices[0].s_length = 2048
 * d_slices[0].s_virt_offset = 0
 * d_slices[0].s_virt_end = 2048	(s_virt_offset + s_length)
 *
 * d_slices[1].s_offset = 4096
 * d_slices[1].s_length = 3072
 * d_slices[1].s_virt_offset = 2048	(d_slices[0].s_virt_end)
 * d_slices[1].s_virt_end = 5120	(s_virt_offset + s_length)
 */
struct g_stripe_slice {
	off_t	s_offset;	/* Offset in source disk. */
	off_t	s_length;	/* Slice length. */
	off_t	s_virt_offset;	/* Offset in destination disk. */
	off_t	s_virt_end;	/* End in destination disk. */
};

/*
 * Structure g_stripe_group describes group of disks that are used
 * on given offset. This allows to create stripe on disks with different
 * sizes.
 *
 *      disk 0     disk 1     disk 2   START_DISK START_OFFSET
 *     +-------+  +-------+  +-------+ 0          0
 *     |0000000|  |0000000|  |0000000|
 *     |0000000|  |0000000|  |0000000|
 *     + - - - +  +-------+  + - - - + 1024       0 + 1024 * 3 = 3072
 *     |1111111|             |1111111|
 *     |1111111|             |1111111|
 *     |1111111|             |1111111|
 *     +-------+             + - - - + 2560       3072 + 1536 * 2 = 6144
 *                           |2222222|
 *                           +-------+ 3072       6144 + 512 * 1 = 6656
 *
 * sc_groups[0].g_ndisks = 3
 * sc_groups[0].g_disks = { 0, 1, 2 }
 * sc_groups[0].g_start_offset = 0
 * sc_groups[0].g_end_offset = 3072
 * sc_groups[0].g_start_disk = 0
 *
 * sc_groups[1].g_ndisks = 2
 * sc_groups[1].g_disks = { 0, 2 }
 * sc_groups[1].g_start_offset = 3072
 * sc_groups[1].g_end_offset = 6144
 * sc_groups[1].g_start_disk = 2560
 *
 * sc_groups[2].g_ndisks = 1
 * sc_groups[2].g_disks = { 2 }
 * sc_groups[2].g_start_offset = 6144
 * sc_groups[2].g_end_offset = 6656
 * sc_groups[2].g_start_disk = 3072
 */
struct g_stripe_group {
	uint16_t	g_ndisks;
	struct g_stripe_disk **g_disks;
	off_t		g_start_offset;
	off_t		g_end_offset;
	off_t		g_start_disk;
};

/*
 * Structure g_stripe_disk describes single disk in stripe.
 */
struct g_stripe_disk {
	struct g_consumer	*d_consumer;
	off_t			 d_size;
	struct g_stripe_slice	*d_slices;
	uint16_t		 d_nslices;
	uint16_t		 d_ndone;
};

struct g_stripe_softc {
	struct g_provider	*sc_provider;

	struct g_stripe_disk	*sc_disks;
	uint16_t		 sc_ndisks;
	uint16_t		 sc_nactive;

	struct g_stripe_group	*sc_groups;
	uint16_t		 sc_ngroups;
	uint16_t		 sc_nactive_slices;

	uint32_t		 sc_sectorsize;
	uint32_t		 sc_stripesize; /* stripe size */
	uint8_t		 	 sc_stripebits; /* stripesize = 2**stripebits */

	void			(*sc_orphan)(struct g_provider *);
	void			*sc_priv;	/* Private data. */
};


struct g_geom *g_stripe_create(struct g_class *mp, const char *name,
    uint16_t ndisks, uint32_t stripesize, uint32_t sectorsize, void *priv,
    void (*orphan)(struct g_provider *));
void g_stripe_add_slice(struct g_geom *gp, struct g_provider *pp, off_t offset,
    off_t length);
int g_stripe_attach(struct g_geom *gp, struct g_provider *pp, uint16_t no,
    uint16_t nslices);
int g_stripe_detach(struct g_geom *gp, struct g_provider *pp);
int g_stripe_destroy(struct g_geom *gp, int force);

#ifndef	BITCOUNT
#define	BITCOUNT(x)	(((BX_(x) + (BX_(x) >> 4)) & 0x0F0F0F0F) % 255)
#define	BX_(x)		((x) - (((x) >> 1) & 0x77777777) -		\
			(((x) >> 2) & 0x33333333) - (((x) >> 3) & 0x11111111))
#endif

#endif	/* _G_STRIPE_H_ */
